#!/usr/bin/env python
# -*- coding: utf-8 -*-

from sklearn.svm import SVC
import logging
from sklearn.model_selection import cross_val_score, GridSearchCV
from sklearn.model_selection import RandomizedSearchCV
from sklearn.ensemble import RandomForestClassifier


logging.basicConfig(level=logging.INFO)


def train(X, Y, args):
    """
    Train a model given the arguments, the dataset and 
    the corresponding labels (ground-truth)

    Parameters
    ----------

    X : array
        features of the dataset
    Y : array
        corresponding labels
    args : dict
        arguments to prepare the model

    Returns
    -------

    model : object
        trained model

    """

    # SVM model selected
    if args.model == "svm":
        logging.info(f"Training SVM model...")

        # Using predefined parameters
        if args.gridsearch == "n":
            logging.info(f"Using predefined parameters.")

            # Training SVM model using radial kernel and predefined parameters
            kernel = "rbf"
            gamma = 0.0001
            C = 1000

            svm_model = SVC(kernel=kernel, gamma=gamma, C=C)
            svm_model.fit(X, Y)

            return svm_model

        # Grid search
        elif args.gridsearch == "y":
            logging.info(f"Doing grid search, it may take a while...")

            # Create the parameter grid
            params_grid = [
                {"kernel": ["rbf"], "gamma": [1e-2, 1e-3, 1e-4], "C": [10, 100, 1000],},
                {"kernel": ["linear"], "C": [10, 100, 1000]},
                {
                    "kernel": ["poly"],
                    "gamma": [1e-2, 1e-3, 1e-4],
                    "degree": [3, 4, 5],
                    "C": [10, 100, 1000],
                },
                {
                    "kernel": ["sigmoid"],
                    "gamma": [1e-2, 1e-3, 1e-4],
                    "C": [10, 100, 1000],
                },
            ]

            svm_model = GridSearchCV(SVC(), params_grid, cv=3, verbose=10, n_jobs=-1)
            svm_model.fit(X, Y)

            logging.info(f"Using hyperparameters: {svm_model.best_params_}")

            return svm_model

    # Random forest model selected
    elif args.model == "rf":
        logging.info(f"Training RF model...")

        # Using predefined parameters
        if args.gridsearch == "n":
            logging.info(f"Using predefined parameters.")

            # Training RF model using predefined parameters
            n_estimators = 50
            max_depth = 25
            min_samples_split = 2
            min_samples_leaf = 4
            bootstrap = True

            rf_model = RandomForestClassifier(
                max_depth=max_depth,
                n_estimators=n_estimators,
                min_samples_split=min_samples_split,
                min_samples_leaf=min_samples_leaf,
                bootstrap=bootstrap,
                random_state=42,
            )

            rf_model.fit(X, Y)

            return rf_model

        # Grid search
        elif args.gridsearch == "y":
            logging.info(f"Doing grid search, it may take a while...")

            n_estimators = [50, 75, 100]
            max_depth = [10, 25, 50]
            min_samples_split = [2, 4, 6]
            min_samples_leaf = [1, 2, 4]
            bootstrap = [True]

            param_grid = {
                "n_estimators": n_estimators,
                "max_depth": max_depth,
                "min_samples_split": min_samples_split,
                "min_samples_leaf": min_samples_leaf,
                "bootstrap": bootstrap,
            }

            rf = RandomForestClassifier(random_state=42)

            rf_model = GridSearchCV(
                estimator=rf, param_grid=param_grid, cv=3, verbose=10, n_jobs=-1
            )
            rf_model.fit(X, Y)

            logging.info(f"Using hyperparameters: {rf_model.best_params_}")

            return rf_model


def predict(X, model):
    """
    Predict labels given the features and the trained model

    Parameters
    ----------

    X : array
        features to predict on
    model : object
        trained model

    Returns
    -------

    predictions : array
        Array with the predicted labels

    """
    Y_pred = model.predict(X)

    return Y_pred



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import logging
import requests, zipfile, io, os

logging.basicConfig(level=logging.INFO)

# URL and Root path of the original data folder
URL = "https://archive.ics.uci.edu/ml/machine-learning-databases/00240/UCI%20HAR%20Dataset.zip"
DATASET_PATH = "UCI HAR Dataset"
TRAIN_DATA = DATASET_PATH + "/train/X_train.txt"
TRAIN_LABELS = DATASET_PATH + "/train/y_train.txt"
TEST_DATA = DATASET_PATH + "/test/X_test.txt"
TEST_LABELS = DATASET_PATH + "/test/y_test.txt"


def download_dataset():
    """
    Download raw dataset from url and unzip it
    """

    if not os.path.isdir(DATASET_PATH) or len(os.listdir(DATASET_PATH)) == 0:

        logging.info(f"Dataset not locally available, downloading...")

        r = requests.get(URL)

        if not r.ok:
            logging.info(f"Error while downloading: {r.status_code}")
            return

        logging.info(f"Extracting raw data...")

        z = zipfile.ZipFile(io.BytesIO(r.content))
        z.extractall()
    else:
        logging.info(f"Dataset already available, skipping download.")


def transform_to_text_labels(labels):
    """
    Transform numerical labels to corresponding text

    Parameters
    ----------

    labels : array
        an array of numerical labels

    Returns
    -------

    labels : array
        same array with corresponding text labels

    """

    labels = labels.astype(str)

    with open(DATASET_PATH + "/activity_labels.txt", "r") as f:
        for row in f:
            num_label, text_label = row.replace("\n", "").split(" ")
            labels = np.where(labels == str(num_label), text_label, labels)

    return labels


def get_dataset_split(data_path, labels_path):
    """
    Get data and ground-truth of selected split

    Parameters
    ----------

    data_path : str
        data file location
    labels_path : str
        labels file location

    Returns
    -------

    data : array
        all the data of the split
    labels: array
        all the corresponding labels (ground-truth)

    """

    # Load data
    with open(data_path, "r") as f:
        data = np.array([row.replace("  ", " ").strip().split(" ") for row in f])

    # Load labels
    with open(labels_path, "r") as f:
        labels = np.array([row.replace("\n", "") for row in f], dtype=int)

    return data, labels


def load(
    standardized=False,
    printSize=False,
    train_data_path=None,
    train_labels_path=None,
    test_data_path=None,
    test_labels_path=None,
):
    """
    Get the dataset and the corresponding labels split
    into a training and a testing set

    Parameters
    ----------

    standardized : bool
        standardize the data before returning them or not

    Returns
    -------

    train_data : array
    train_labels: array
    test_data : array
    test_labels: array

    """

    logging.info(f"Starting dataset loading...")

    download_dataset()

    # Get training data
    if train_data_path and train_labels_path:
        train_data, train_labels = get_dataset_split(train_data_path, train_labels_path)
        logging.info(f"Custom Train Dataset loaded.")
    else:
        train_data, train_labels = get_dataset_split(TRAIN_DATA, TRAIN_LABELS)

    # Get testing data
    if test_data_path and test_labels_path:
        test_data, test_labels = get_dataset_split(test_data_path, test_labels_path)
        logging.info(f"Custom Test Dataset loaded.")
    else:
        test_data, test_labels = get_dataset_split(TEST_DATA, TEST_LABELS)

    logging.info(f"Dataset ready.")

    if printSize:
        logging.info(f"---Train samples: {train_data.shape[0]}")
        logging.info(f"---Test samples: {test_data.shape[0]}")

    # Standardization if required
    if standardized:
        from .preprocessor import standardize

        train_data, test_data = standardize(train_data, test_data)

    return train_data, train_labels, test_data, test_labels



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pandas as pd
import argparse
import logging
import sys

from . import database
from . import algorithm
import argparse
from . import evaluator

logging.basicConfig(level=logging.INFO)


def main():
    args = get_args()
    logging.info(f"Using arguments: {args}")

    # Load data and ground-truth
    train_data, train_labels, test_data, test_labels = database.load(
        standardized=True,
        printSize=True,
        train_data_path=args.train_data,
        train_labels_path=args.train_labels,
        test_data_path=args.test_data,
        test_labels_path=args.test_labels,
    )
    train_labels = train_labels.ravel()

    # Training
    model = algorithm.train(train_data, train_labels, args)

    # Logging scores on train and test sets
    logging.info("---Training set accuracy: %f" % model.score(train_data, train_labels))
    logging.info("---Testing  set accuracy: %f" % model.score(test_data, test_labels))

    predictions = algorithm.predict(test_data, model)

    # Evaluate the predictions
    evaluator.evaluate(
        predictions, test_data, test_labels, args.output_folder, args.model, model
    )


def get_args(args=None):
    parser = argparse.ArgumentParser(
        """Train using a Support Vector Machine (SVM) model or a Random Forest 
        (RF) model. You can also train with (very slow) or without doing 
        Grid Search for performing hyper parameter tuning 
        
        Example: python run.py -model rf -gridsearch n -output-folder results
        """
    )
    parser.add_argument(
        "-model",
        type=str,
        default="svm",
        help="Use SVM or RF ? --> svm [default] or rf",
        dest="model",
        choices=["svm", "rf"],
    )
    parser.add_argument(
        "-gridsearch",
        type=str,
        default="n",
        help="Do Grid Search ? --> y or n [default]",
        dest="gridsearch",
        choices=["y", "n"],
    )
    parser.add_argument(
        "-output-folder",
        type=str,
        default="results",
        help="Path where to store the evaluation results (created if does not exist)",
        dest="output_folder",
    )
    parser.add_argument(
        "-train-data",
        type=str,
        default=None,
        help="Path to custom train_data.txt file",
        dest="train_data",
    )
    parser.add_argument(
        "-train-labels",
        type=str,
        default=None,
        help="Path to custom train_labels.txt file",
        dest="train_labels",
    )
    parser.add_argument(
        "-test-data",
        type=str,
        default=None,
        help="Path to custom test_data.txt file",
        dest="test_data",
    )
    parser.add_argument(
        "-test-labels",
        type=str,
        default=None,
        help="Path to custom test_labels.txt file",
        dest="test_labels",
    )

    return parser.parse_args(args)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys
import numpy as np
import pytest
from . import database
from . import evaluator
from . import algorithm
from . import run
from . import preprocessor

import logging

model = None
train_data, train_labels, test_data, test_labels = None, None, None, None

# ========================================================================

"""Tests the database script"""


def test_download_dataset():

    database.download_dataset()

    assert os.path.exists("UCI HAR Dataset")
    assert os.path.isdir("UCI HAR Dataset")

    # Check if the right files and folders are present
    entries = os.listdir("UCI HAR Dataset")
    expected_files = [
        "activity_labels.txt",
        "features_info.txt",
        "features.txt",
        "test",
        "train",
    ]
    for f in expected_files:
        assert f in entries


def test_transform_to_text_labels():

    num_labels = np.array([1, 2, 3, 4, 5, 6])
    labels = database.transform_to_text_labels(num_labels)

    assert np.array_equal(
        labels,
        np.array(
            [
                "WALKING",
                "WALKING_UPSTAIRS",
                "WALKING_DOWNSTAIRS",
                "SITTING",
                "STANDING",
                "LAYING",
            ]
        ),
    )


def test_get_dataset_split():
    train_data = "UCI HAR Dataset/train/X_train.txt"
    train_labels = "UCI HAR Dataset/train/y_train.txt"

    train_data, train_labels = database.get_dataset_split(train_data, train_labels)

    assert train_data.shape == (7352, 561)
    assert train_labels.shape == (7352,)
    assert min(train_labels) == 1
    assert max(train_labels) == 6

    test_data = "UCI HAR Dataset/test/X_test.txt"
    test_labels = "UCI HAR Dataset/test/y_test.txt"

    test_data, test_labels = database.get_dataset_split(test_data, test_labels)

    assert test_data.shape == (2947, 561)
    assert test_labels.shape == (2947,)
    assert min(test_labels) == 1
    assert max(test_labels) == 6


def test_load(caplog):

    caplog.set_level(logging.INFO)

    # Save data for other tests
    (
        pytest.train_data,
        pytest.train_labels,
        pytest.test_data,
        pytest.test_labels,
    ) = database.load(standardized=True, printSize=True)

    assert caplog.record_tuples[3][2] == "---Train samples: 7352"
    assert caplog.record_tuples[4][2] == "---Test samples: 2947"
    assert caplog.record_tuples[5][2] == "Dataset standardized."

    # Save data for other tests
    __, __, __, __ = database.load(
        standardized=True,
        printSize=True,
        train_data_path="UCI HAR Dataset/train/X_train.txt",
        train_labels_path="UCI HAR Dataset/train/y_train.txt",
        test_data_path="UCI HAR Dataset/test/X_test.txt",
        test_labels_path="UCI HAR Dataset/test/y_test.txt",
    )

    assert caplog.record_tuples[3][2] == "---Train samples: 7352"
    assert caplog.record_tuples[4][2] == "---Test samples: 2947"
    assert caplog.record_tuples[5][2] == "Dataset standardized."


# ========================================================================

"""Tests the preprocessor script"""


def test_standardize():
    train_data = np.array([[7, 5, 2, 6], [1, 8, 5, 1], [2, 7, 2, 1]])
    train_mean = np.mean(train_data, axis=0)
    train_std = np.std(train_data, axis=0)
    train_transfo = (train_data - train_mean) / train_std

    train_data_std, test_data_std = preprocessor.standardize(train_data, train_data)

    assert np.array_equal(train_transfo, test_data_std)
    assert np.array_equal(train_transfo, train_data_std)


# ========================================================================

"""Tests the algorithm script"""


def test_algorithm():
    args = run.get_args(["-model", "rf"])
    args_svm = run.get_args(["-model", "svm"])

    pytest.model = algorithm.train(pytest.train_data, pytest.train_labels, args)
    pytest.model_2 = algorithm.train(pytest.train_data, pytest.train_labels, args_svm)

    assert pytest.model.get_params().get("n_estimators") == 50
    assert pytest.model.get_params().get("max_depth") == 25
    assert pytest.model.get_params().get("min_samples_split") == 2
    assert pytest.model.get_params().get("min_samples_leaf") == 4
    assert pytest.model.get_params().get("bootstrap") == True
    assert type(pytest.model).__name__ == "RandomForestClassifier"

    assert pytest.model_2.get_params().get("kernel") == "rbf"
    assert pytest.model_2.get_params().get("gamma") == 0.0001
    assert pytest.model_2.get_params().get("C") == 1000
    assert type(pytest.model_2).__name__ == "SVC"


def test_predict():
    pytest.predictions = algorithm.predict(pytest.test_data, pytest.model)

    assert pytest.predictions[0] == 5
    assert pytest.predictions[50] == 5
    assert pytest.predictions[100] == 1
    assert pytest.predictions[-1] == 1


# ========================================================================

"""Tests the evaluator script"""


def test_get_metrics_table():
    # Fake data
    predictedLabels = np.array([0, 1, 2, 3, 4, 5])
    trueLabels = np.array([0, 1, 2, 1, 2, 3])

    table = evaluator.get_metrics_table(predictedLabels, trueLabels)

    # Check if we get the correct metrics
    assert table.count("0.5") == 4
    assert "Precision" in table
    assert "Recall" in table
    assert "F1 score" in table
    assert "Accuracy" in table


def test_get_table_header():

    table = evaluator.get_table_header("rf", pytest.model)

    assert "Model used: rf" in table
    assert "Parameters:" in table
    assert len(table.splitlines()) == 7


def test_evaluate(caplog):
    caplog.set_level(logging.INFO)

    evaluator.evaluate(
        pytest.predictions,
        pytest.test_data,
        pytest.test_labels,
        "results",
        "rf",
        pytest.model,
    )

    assert "Saving table at" in caplog.record_tuples[1][2]
    assert "Saving confusion matrix at" in caplog.record_tuples[2][2]

    assert os.path.isfile(os.getcwd() + "/results/table.rst")
    assert os.path.isfile(os.getcwd() + "/results/confusion_matrix.png")


# ========================================================================

"""Tests the main script"""


def test_get_args():
    args = run.get_args(["-model", "rf"])

    assert args.gridsearch == "n"
    assert args.model == "rf"
    assert args.output_folder == "results"


def test_main_function(caplog):
    caplog.set_level(logging.INFO)

    sys.argv = ["main", "-model", "rf", "-gridsearch", "n", "-output-folder", "results"]

    run.main()

    assert "Dataset ready." in caplog.messages
    assert "Training RF model..." in caplog.messages
    assert "Starting evaluation..." in caplog.messages



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sklearn
import tabulate
import os
import logging
import numpy as np
import matplotlib.pyplot as plt
from .database import transform_to_text_labels
from sklearn.utils.multiclass import unique_labels
from sklearn.metrics import (
    precision_score,
    recall_score,
    f1_score,
    accuracy_score,
    confusion_matrix,
    classification_report,
)
from sklearn.metrics import plot_confusion_matrix


def get_metrics_table(predictions, test_labels):
    """
    Generate a metrics table to evaluate predictions

    Parameters
    ----------

    predictions : array
        Predictions of a model
    test_labels : array
        Corresponding ground-truth

    Returns
    -------

    table : string
        Nicely formatted plain-text table with the computed metrics

    """

    # Compute metrics
    precision = precision_score(test_labels, predictions, average="micro")
    recall = recall_score(test_labels, predictions, average="micro")
    f1 = f1_score(test_labels, predictions, average="micro")
    accuracy = accuracy_score(test_labels, predictions)

    # Create table
    headers = ["Precision (avg)", "Recall (avg)", "F1 score (avg)", "Accuracy"]

    table = [[precision, recall, f1, accuracy]]

    return tabulate.tabulate(table, headers, tablefmt="rst", floatfmt=".3f")


def get_table_header(model_name, model):
    """
    Generate a header for the metrics table

    Parameters
    ----------

    model_name : str
        Type of model (svm or rf)
    model : object
        Trained model from which to get the parameters

    Returns
    -------

    header : string
        Nicely formatted text header

    """

    header = f"Model used: {str(model_name)}\n" + "Parameters:\n"

    if type(model) is sklearn.model_selection._search.GridSearchCV:
        model = model.best_estimator_

    if model_name == "svm":
        header += (
            f"-kernel: {model.kernel}\n"
            + f"-gamma: {model.gamma}\n"
            + f"-C: {model.C}\n"
        )
    elif model_name == "rf":
        header += (
            f"-n_estimators: {model.n_estimators}\n"
            + f"-max_depth: {model.max_depth}\n"
            + f"-min_samples_split: {model.min_samples_split}\n"
            + f"-min_samples_leaf: {model.min_samples_leaf}\n"
            + f"-bootstrap: {model.bootstrap}\n"
        )

    return header


def evaluate(predictions, test_data, test_labels, output_dir, model_name, model):
    """
    Evaluate the predictions given the ground-truth. Save a table with
    the metrics and a png file with the confusion matrix.

    Parameters
    ----------

    predictions : array
        Predictions of a model
    test_labels : array
        Corresponding ground-truth
    output_dir : str
        Folder name in which to save table and figure
    model_name : str
        Model type (svm or rf)
    model : object
        trained model

    """

    logging.info(f"Starting evaluation...")

    # Get the metrics table
    table = get_metrics_table(predictions, test_labels)

    # Append header to the table
    table = get_table_header(model_name, model) + table

    # Save the metrics table
    output_table = os.getcwd() + "/" + output_dir + "/table.rst"
    logging.info(f"Saving table at {output_table}")
    os.makedirs(os.path.dirname(output_table), exist_ok=True)
    with open(output_table, "wt") as f:
        f.write(table)

    # Numerical labels to text for the plot
    test_labels_txt = set(test_labels)
    classes_txt = transform_to_text_labels(np.array(list(test_labels_txt)))

    fig, ax = plt.subplots(figsize=(17, 15))

    # Get the normalized confusion matrix
    fig = plot_confusion_matrix(
        model,
        test_data,
        test_labels,
        normalize="true",
        cmap=plt.cm.Blues,
        display_labels=classes_txt,
        xticks_rotation="vertical",
        ax=ax,
    )

    fig.ax_.set(ylabel="True label", xlabel="Predicted label")

    title = f"Normalized confusion matrix\nModel: {model_name}"
    fig.ax_.set_title(title)

    # Save the confusion matrix
    output_figure = os.getcwd() + "/" + output_dir + "/confusion_matrix.png"
    logging.info(f"Saving confusion matrix at {output_figure}")
    plt.savefig(output_figure)



#!/usr/bin/env python
# coding=utf-8

from setuptools import setup, find_packages


def load_requirements(f):
    retval = [str(k.strip()) for k in open(f, "rt")]
    return [k for k in retval if k and k[0] not in ("#", "-")]


setup(
    name="rrgp",
    version="1.0.6",
    description="This project aims to train a model to recognise human activities (like walking, standing, or sitting) based on accelerometer and gyroscope data collected with a smartphone.",
    url="https://github.com/patrickaudriaz/mini-project",
    license="MIT",
    author="Geoffrey Raposo, Patrick Audriaz",
    author_email="geoffrey@raposo.ch, audriazp@gmail.com",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    packages=find_packages(),
    include_package_data=True,
    install_requires=load_requirements("requirements.txt"),
    entry_points={"console_scripts": ["rrgp = rrgp.run:main"]},
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from sklearn.preprocessing import StandardScaler
import logging


def standardize(train_data, test_data):
    """
    Standardize training and testing data

    Parameters
    ----------

    train_data : array
        Data on which to calculate the standardization parameters.
        The standardization is also applied on this subset.
    test_data : array
        Test subset on which to apply the standardization.

    Returns
    -------

    train_data : array
        Standardized training data
    test_data : array
        Standardized testing data

    """

    # Instantiate scikit standard scaler
    scaler = StandardScaler()

    # Fit and standardize on training data
    train_data_std = scaler.fit_transform(train_data)

    # Standardize testing data
    test_data_std = scaler.transform(test_data)

    logging.info(f"Dataset standardized.")

    return train_data_std, test_data_std
